from typing import Dict

import numpy as np
import pandas as pd

from util.exceptions import ImplicitMetricValueError, InvalidMetricValue
from util.get_functions import create_coordinate_pair


def interpolate_series(series, x, y):
    """
    Interpolates a series using the given x and y values.

    This function interpolates a series using the given x and y values.
    """

    return [np.interp(item / 100, x, y) for item in series]


def resolve_metric_list_parameter(metric):
    """
    Resolves the metric list parameter to calculate a measure

    This functions converts the metric parameter to a pandas Series if it is a list (CalculateMeasures endpoint)
    otherwise it just returns the metric - already a pandas Series (Analysis endpoint).
    """
    return pd.Series(metric, dtype=np.float64) if isinstance(metric, list) else metric


def calculate_em1(data: Dict):
    """
    Calculates non-complex files density (em1).

    This function calculates non-complex files density measure (em1)
    used to assess the changeability quality sub characteristic.
    """
    files_complexity = resolve_metric_list_parameter(data["complexity"])
    files_functions = resolve_metric_list_parameter(data["functions"])

    if "number_of_files" in data:
        number_of_files = data["number_of_files"]

    elif len(files_complexity) != len(files_functions):
        raise ImplicitMetricValueError((
            "Unable to get the implicit value of metric `number_of_file` "
            "because the size of the lists of values of metrics `complexity` "
            "and `functions` are not equal."
        ))
    else:
        number_of_files = len(files_complexity)

    has_none = files_complexity is None or files_functions is None
    has_zero = len(files_complexity) == 0 or len(files_functions) == 0

    if has_none or has_zero:
        return 0

    COMPLEX_FILES_DENSITY_THRESHOLD = 10

    if files_complexity.sum() <= 0:
        raise InvalidMetricValue(
            "The cyclomatic complexity of all files is lesser or equal than 0"
        )

    if files_functions.sum() <= 0:
        raise InvalidMetricValue(
            "The number of functions of all files is lesser or equal than 0"
        )

    x, y = create_coordinate_pair(0, COMPLEX_FILES_DENSITY_THRESHOLD)

    files_in_thresholds_df = (files_complexity / files_functions) <= COMPLEX_FILES_DENSITY_THRESHOLD
    IF1 = np.interp(list(files_in_thresholds_df[(files_functions > 0)]), x, y)
    em1 = sum(IF1) / number_of_files
    return em1


def calculate_em2(data: Dict):
    """
    Calculates commented files density (em2).

    This function calculates commented files density measure (em2)
    used to assess the changeability quality sub characteristic.
    """
    files_comment_lines_density = resolve_metric_list_parameter(
        data["comment_lines_density"]
    )

    if "number_of_files" in data:
        number_of_files = data["number_of_files"]
    else:
        number_of_files = len(files_comment_lines_density)

    has_none = files_comment_lines_density is None
    has_zero = len(files_comment_lines_density) == 0

    if has_none or has_zero:
        return 0

    MINIMUM_COMMENT_DENSITY_THRESHOLD = 10
    MAXIMUM_COMMENT_DENSITY_THRESHOLD = 30

    if files_comment_lines_density.sum() < 0:
        raise InvalidMetricValue(
            "The number of files comment lines density is lesser than 0"
        )

    x, y = create_coordinate_pair(
        MINIMUM_COMMENT_DENSITY_THRESHOLD / 100, MAXIMUM_COMMENT_DENSITY_THRESHOLD / 100
    )

    files_between_thresholds = files_comment_lines_density[
        files_comment_lines_density.between(
            MINIMUM_COMMENT_DENSITY_THRESHOLD,
            MAXIMUM_COMMENT_DENSITY_THRESHOLD,
            inclusive="both",
        )
    ]

    em2i = interpolate_series(files_between_thresholds, x, y)
    em2 = np.sum(em2i) / number_of_files
    return em2


def calculate_em3(data: Dict):
    """
    Calculates duplicated files absence (em3).

    This function calculates the duplicated files absence measure (em3)
    used to assess the changeability quality sub characteristic.
    """
    files_duplicated_lines_density = resolve_metric_list_parameter(
        data["duplicated_lines_density"]
    )

    if "number_of_files" in data:
        number_of_files = data["number_of_files"]
    else:
        number_of_files = len(files_duplicated_lines_density)

    has_none = files_duplicated_lines_density is None
    has_zero = len(files_duplicated_lines_density) == 0

    if has_none or has_zero:
        return 0

    DUPLICATED_LINES_THRESHOLD = 5.0

    if files_duplicated_lines_density.sum() < 0:
        raise InvalidMetricValue(
            "The number of files duplicated lines density is lesser than 0"
        )

    x, y = create_coordinate_pair(0, DUPLICATED_LINES_THRESHOLD / 100)

    files_below_threshold = files_duplicated_lines_density[
        files_duplicated_lines_density <= DUPLICATED_LINES_THRESHOLD
    ]

    em3i = interpolate_series(files_below_threshold, x, y)
    em3 = np.sum(em3i) / number_of_files
    return em3


def calculate_em4(data: Dict[str, float]):
    """
    Calculates passed tests (em4)

    This function calculates the passed tests measure (em4)
    used to assess the testing status sub characteristic.
    """
    try:
        number_of_tests = resolve_metric_list_parameter(data["tests"]).sum()
        number_of_test_errors = data["test_errors"]
        number_of_test_failures = data["test_failures"]

        x, y = create_coordinate_pair(0, 1, reverse_y=True)

        number_of_fail_tests = number_of_test_errors + number_of_test_failures
        if4i = (number_of_tests - number_of_fail_tests) / number_of_tests
        return np.interp(if4i, x, y)
    except ZeroDivisionError:
        return 0


def calculate_em5(data: Dict[str, list]):
    """
    Calculates fast test builds (em5)

    This function calculates the fast test builds measure (em5)
    used to assess the testing status sub characteristic.
    """
    execution_time = resolve_metric_list_parameter(data["test_execution_time"])
    number_of_tests = resolve_metric_list_parameter(data["tests"])
    number_of_files = len(execution_time)

    has_none = execution_time is None or number_of_tests is None
    has_zero = len(execution_time) == 0 or len(number_of_tests) == 0

    if has_none or has_zero:
        return 0

    MAXIMUM_COVERAGE_THRESHOLD = 300000

    x, y = create_coordinate_pair(
        0,
        MAXIMUM_COVERAGE_THRESHOLD
    )

    execution_between_thresholds = execution_time[execution_time <= MAXIMUM_COVERAGE_THRESHOLD]
    fast_tests_between_thresholds = execution_between_thresholds / number_of_tests

    em5i = interpolate_series(fast_tests_between_thresholds, x, y)
    em5 = np.sum(em5i) / number_of_files
    return em5


def calculate_em6(data: Dict):
    """
    Calculates test coverage (em6).

    This function calculates the test coverage measure (em6)
    used to assess the testing status sub characteristic.
    """
    coverage = resolve_metric_list_parameter(data["coverage"])

    if "number_of_files" in data:
        number_of_files = data["number_of_files"]
    else:
        number_of_files = len(coverage)

    has_none = coverage is None
    has_zero = len(coverage) == 0

    if has_none or has_zero:
        return 0

    MINIMUM_COVERAGE_THRESHOLD = 60
    MAXIMUM_COVERAGE_THRESHOLD = 90

    x, y = create_coordinate_pair(
        MINIMUM_COVERAGE_THRESHOLD / 100,
        MAXIMUM_COVERAGE_THRESHOLD / 100,
        reverse_y=True,
    )

    files_between_thresholds = coverage[coverage >= MINIMUM_COVERAGE_THRESHOLD]
    em6i = interpolate_series(files_between_thresholds, x, y)
    em6 = np.sum(em6i) / number_of_files
    return em6


def calculate_em7(data: Dict):
    """
    Calculates team throughput (em7).

    This function calculates the team throughput measure (em7)
    used to assess the functional completeness subcharacteristic.
    """
    resolved_issues_with_us_label = data[
        "number_of_resolved_issues_with_US_label_in_the_last_x_days"
    ]

    total_issues_with_us_label = data[
        "total_number_of_issues_with_US_label_in_the_last_x_days"
    ]

    x, y = create_coordinate_pair(0, 1, reverse_y=True)

    if7 = np.divide(resolved_issues_with_us_label, total_issues_with_us_label)

    return np.interp(if7, x, y)



from marshmallow import Schema, fields, validate


class MeasureSchema(Schema):
    """
    {
        "key": "passed_tests",
        "parameters": {
            "tests": 10,
            "test_errors": 3,
            "test_failures": 1
        }
    }
    """
    key = fields.Str(required=True)
    parameters = fields.Dict(required=True)


class CalculateMeasureSchema(Schema):
    """
    {
        "measures": [
            {
                "key": "passed_tests",
                "parameters": {
                    "tests": 10,
                    "test_errors": 3,
                    "test_failures": 1
                }
            },
            {
                "key": "test_builds",
                "parameters": {
                    "param1": 8,
                    "param2": 19,
                    "parma3": 4
                }
            }
        ]
    }
    """
    measures = fields.List(fields.Nested(MeasureSchema), required=True)


class CalculatedSubEntitySchema(Schema):
    key = fields.Str(required=True)
    value = fields.Float(validate=validate.Range(min=0, max=1), required=True)
    weight = fields.Float(validate=validate.Range(min=0, max=100), required=True)


class SubCharacteristicSchema(Schema):
    key = fields.Str(required=True)
    measures = fields.List(fields.Nested(CalculatedSubEntitySchema), required=True)


class CalculateSubCharacteristicSchema(Schema):
    """
    {
        "subcharacteristics": [
            {
                "key": "testing_status",
                "measures": [
                    {
                        "key": "passed_tests",
                        "value": 1.0,
                        "weight": 33,
                    },
                    ...
                ]
            },
            ...
        ]
    }
    """
    subcharacteristics = fields.List(fields.Nested(SubCharacteristicSchema), required=True)


class CharacteristicSchema(Schema):
    key = fields.Str(required=True)
    subcharacteristics = fields.List(fields.Nested(CalculatedSubEntitySchema), required=True)


class CalculateCharacteristicSchema(Schema):
    """
    {
        "characteristics": [
            {
                "key": "reliability",
                "subcharacteristics": [
                    {
                        "key": "testing_status",
                        "value": 1.0,
                        "weight": 50,
                    },
                    ...
                ]
            },
            ...
        ]
    }
    """
    characteristics = fields.List(fields.Nested(CharacteristicSchema), required=True)


class SQCSchema(Schema):
    key = fields.Str(required=True)
    characteristics = fields.List(fields.Nested(CalculatedSubEntitySchema), required=True)


class CalculateSQCSchema(Schema):
    """
    {
        "sqc": {
            "key": "sqc",
            "characteristics": [
                {
                    "key": "reliability",
                    "value": 1.0,
                    "weight": 50,
                },
                ...
            ]
        }
    }
    """
    sqc = fields.Nested(SQCSchema, required=True)


class NonComplexFileDensitySchema(Schema):
    """
    "key": "non_complex_file_density",
    "function": calculate_em1
    """
    complexity = fields.List(fields.Float(required=True))
    functions = fields.List(fields.Float(required=True))


class CommentedFileDensitySchema(Schema):
    """
    "key": "commented_file_density",
    "function": calculate_em2
    """
    comment_lines_density = fields.List(fields.Float(required=True))


class DuplicationAbsenceSchema(Schema):
    """
    "key": "duplication_absense",
    "function": calculate_em3
    """
    duplicated_lines_density = fields.List(fields.Float(required=True))


class PassedTestsSchema(Schema):
    """
    "key": "passed_tests",
    "function": calculate_em4
    """
    tests = fields.List(fields.Float(required=True))
    test_errors = fields.Float(required=True)
    test_failures = fields.Float(required=True)


class TestBuildsSchema(Schema):
    """
    "key": "test_builds",
    "function": calculate_em5
    """
    test_execution_time = fields.List(fields.Float(required=True))
    tests = fields.List(fields.Float(required=True))


class TestCoverageSchema(Schema):
    """
    "key": "test_coverage",
    "function": calculate_em6
    """
    coverage = fields.List(fields.Float(required=True))


class TeamThroughputSchema(Schema):
    """
    "key": "team_throughput",
    "function": calculate_em7
    """
    number_of_resolved_issues_with_US_label_in_the_last_x_days = fields.Integer(required=True)
    total_number_of_issues_with_US_label_in_the_last_x_days = fields.Integer(required=True)



# flake8: noqa
# pylint: skip-file
SONARQUBE_AVAILABLE_METRICS = {
    "metrics": [
        {
            "id": "289",
            "key": "new_technical_debt",
            "type": "WORK_DUR",
            "name": "Added Technical Debt",
            "description": "Added technical debt",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "386",
            "key": "analysis_from_sonarqube_9_4",
            "type": "BOOL",
            "name": "Analysis From SonarQube 9.4",
            "description": "Indicates whether the analysis has been run after the upgrade to SonarQube 9.4. It affects how the issues will be detected for branches that use reference branch as the strategy for detecting new code.",
            "domain": "Issues",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "78",
            "key": "blocker_violations",
            "type": "INT",
            "name": "Blocker Issues",
            "description": "Blocker issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "308",
            "key": "bugs",
            "type": "INT",
            "name": "Bugs",
            "description": "Bugs",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "4",
            "key": "classes",
            "type": "INT",
            "name": "Classes",
            "description": "Classes",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "306",
            "key": "code_smells",
            "type": "INT",
            "name": "Code Smells",
            "description": "Code Smells",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "347",
            "key": "cognitive_complexity",
            "type": "INT",
            "name": "Cognitive Complexity",
            "description": "Cognitive complexity",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "32",
            "key": "comment_lines",
            "type": "INT",
            "name": "Comment Lines",
            "description": "Number of comment lines",
            "domain": "Size",
            "direction": 1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "239",
            "key": "comment_lines_data",
            "type": "DATA",
            "name": "comment_lines_data",
            "domain": "Size",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "33",
            "key": "comment_lines_density",
            "type": "PERCENT",
            "name": "Comments (%)",
            "description": "Comments balanced by ncloc + comment lines",
            "domain": "Size",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "27",
            "key": "class_complexity",
            "type": "FLOAT",
            "name": "Complexity / Class",
            "description": "Complexity average by class",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": True,
            "hidden": True,
            "decimalScale": 1
        },
        {
            "id": "38",
            "key": "file_complexity",
            "type": "FLOAT",
            "name": "Complexity / File",
            "description": "Complexity average by file",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": True,
            "hidden": True,
            "decimalScale": 1
        },
        {
            "id": "28",
            "key": "function_complexity",
            "type": "FLOAT",
            "name": "Complexity / Function",
            "description": "Complexity average by function",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": True,
            "hidden": True,
            "decimalScale": 1
        },
        {
            "id": "272",
            "key": "complexity_in_classes",
            "type": "INT",
            "name": "Complexity in Classes",
            "description": "Cyclomatic complexity in classes",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "273",
            "key": "complexity_in_functions",
            "type": "INT",
            "name": "Complexity in Functions",
            "description": "Cyclomatic complexity in functions",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "97",
            "key": "branch_coverage",
            "type": "PERCENT",
            "name": "Condition Coverage",
            "description": "Condition coverage",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "178",
            "key": "new_branch_coverage",
            "type": "PERCENT",
            "name": "Condition Coverage on New Code",
            "description": "Condition coverage of new/changed code",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "95",
            "key": "conditions_to_cover",
            "type": "INT",
            "name": "Conditions to Cover",
            "description": "Conditions to cover",
            "domain": "Coverage",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "176",
            "key": "new_conditions_to_cover",
            "type": "INT",
            "name": "Conditions to Cover on New Code",
            "description": "Conditions to cover on new code",
            "domain": "Coverage",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "277",
            "key": "confirmed_issues",
            "type": "INT",
            "name": "Confirmed Issues",
            "description": "Confirmed issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "20",
            "key": "coverage",
            "type": "PERCENT",
            "name": "Coverage",
            "description": "Coverage by tests",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "172",
            "key": "new_coverage",
            "type": "PERCENT",
            "name": "Coverage on New Code",
            "description": "Coverage of new/changed code",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "79",
            "key": "critical_violations",
            "type": "INT",
            "name": "Critical Issues",
            "description": "Critical issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "26",
            "key": "complexity",
            "type": "INT",
            "name": "Cyclomatic Complexity",
            "description": "Cyclomatic complexity",
            "domain": "Complexity",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "181",
            "key": "last_commit_date",
            "type": "MILLISEC",
            "name": "Date of Last Commit",
            "domain": "SCM",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "280",
            "key": "development_cost",
            "type": "STRING",
            "name": "Development Cost",
            "description": "Development cost",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": True
        },
        {
            "id": "348",
            "key": "new_development_cost",
            "type": "FLOAT",
            "name": "Development Cost on New Code",
            "description": "Development cost on new code",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": True,
            "decimalScale": 1
        },
        {
            "id": "37",
            "key": "directories",
            "type": "INT",
            "name": "Directories",
            "description": "Directories",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "51",
            "key": "duplicated_blocks",
            "type": "INT",
            "name": "Duplicated Blocks",
            "description": "Duplicated blocks",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "339",
            "key": "new_duplicated_blocks",
            "type": "INT",
            "name": "Duplicated Blocks on New Code",
            "description": "Duplicated blocks on new code",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "52",
            "key": "duplicated_files",
            "type": "INT",
            "name": "Duplicated Files",
            "description": "Duplicated files",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "8",
            "key": "duplicated_lines",
            "type": "INT",
            "name": "Duplicated Lines",
            "description": "Duplicated lines",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "34",
            "key": "duplicated_lines_density",
            "type": "PERCENT",
            "name": "Duplicated Lines (%)",
            "description": "Duplicated lines balanced by statements",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "340",
            "key": "new_duplicated_lines_density",
            "type": "PERCENT",
            "name": "Duplicated Lines (%) on New Code",
            "description": "Duplicated lines (%) on new code balanced by statements",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "338",
            "key": "new_duplicated_lines",
            "type": "INT",
            "name": "Duplicated Lines on New Code",
            "description": "Duplicated Lines on New Code",
            "domain": "Duplications",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "53",
            "key": "duplications_data",
            "type": "DATA",
            "name": "Duplication Details",
            "description": "Duplications details",
            "domain": "Duplications",
            "direction": 0,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "312",
            "key": "effort_to_reach_maintainability_rating_a",
            "type": "WORK_DUR",
            "name": "Effort to Reach Maintainability Rating A",
            "description": "Effort to reach maintainability rating A",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "321",
            "key": "executable_lines_data",
            "type": "DATA",
            "name": "executable_lines_data",
            "domain": "Coverage",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "274",
            "key": "False_positive_issues",
            "type": "INT",
            "name": "False Positive Issues",
            "description": "False positive issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "154",
            "key": "file_complexity_distribution",
            "type": "DISTRIB",
            "name": "File Distribution / Complexity",
            "description": "Files distribution /complexity",
            "domain": "Complexity",
            "direction": 0,
            "qualitative": True,
            "hidden": True
        },
        {
            "id": "36",
            "key": "files",
            "type": "INT",
            "name": "Files",
            "description": "Number of files",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "59",
            "key": "function_complexity_distribution",
            "type": "DISTRIB",
            "name": "Function Distribution / Complexity",
            "description": "Functions distribution /complexity",
            "domain": "Complexity",
            "direction": 0,
            "qualitative": True,
            "hidden": True
        },
        {
            "id": "6",
            "key": "functions",
            "type": "INT",
            "name": "Functions",
            "description": "Functions",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "150",
            "key": "generated_lines",
            "type": "INT",
            "name": "Generated Lines",
            "description": "Number of generated lines",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "151",
            "key": "generated_ncloc",
            "type": "INT",
            "name": "Generated Lines of Code",
            "description": "Generated non Commenting Lines of Code",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "82",
            "key": "info_violations",
            "type": "INT",
            "name": "Info Issues",
            "description": "Info issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "42",
            "key": "violations",
            "type": "INT",
            "name": "Issues",
            "description": "Issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "94",
            "key": "line_coverage",
            "type": "PERCENT",
            "name": "Line Coverage",
            "description": "Line coverage",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "175",
            "key": "new_line_coverage",
            "type": "PERCENT",
            "name": "Line Coverage on New Code",
            "description": "Line coverage of added/changed code",
            "domain": "Coverage",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "55",
            "key": "lines",
            "type": "INT",
            "name": "Lines",
            "description": "Lines",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "5",
            "key": "ncloc",
            "type": "INT",
            "name": "Lines of Code",
            "description": "Non commenting lines of code",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "293",
            "key": "ncloc_language_distribution",
            "type": "DATA",
            "name": "Lines of Code Per Language",
            "description": "Non Commenting Lines of Code Distributed By Language",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "92",
            "key": "lines_to_cover",
            "type": "INT",
            "name": "Lines to Cover",
            "description": "Lines to cover",
            "domain": "Coverage",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "173",
            "key": "new_lines_to_cover",
            "type": "INT",
            "name": "Lines to Cover on New Code",
            "description": "Lines to cover on new code",
            "domain": "Coverage",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "163",
            "key": "sqale_rating",
            "type": "RATING",
            "name": "Maintainability Rating",
            "description": "A-to-E rating based on the technical debt ratio",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "341",
            "key": "new_maintainability_rating",
            "type": "RATING",
            "name": "Maintainability Rating on New Code",
            "description": "Maintainability rating on new code",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "80",
            "key": "major_violations",
            "type": "INT",
            "name": "Major Issues",
            "description": "Major issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "81",
            "key": "minor_violations",
            "type": "INT",
            "name": "Minor Issues",
            "description": "Minor issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "238",
            "key": "ncloc_data",
            "type": "DATA",
            "name": "ncloc_data",
            "domain": "Size",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "165",
            "key": "new_blocker_violations",
            "type": "INT",
            "name": "New Blocker Issues",
            "description": "New Blocker issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "309",
            "key": "new_bugs",
            "type": "INT",
            "name": "New Bugs",
            "description": "New Bugs",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "307",
            "key": "new_code_smells",
            "type": "INT",
            "name": "New Code Smells",
            "description": "New Code Smells",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "166",
            "key": "new_critical_violations",
            "type": "INT",
            "name": "New Critical Issues",
            "description": "New Critical issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "169",
            "key": "new_info_violations",
            "type": "INT",
            "name": "New Info Issues",
            "description": "New Info issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "164",
            "key": "new_violations",
            "type": "INT",
            "name": "New Issues",
            "description": "New issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "337",
            "key": "new_lines",
            "type": "INT",
            "name": "New Lines",
            "description": "New lines",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "167",
            "key": "new_major_violations",
            "type": "INT",
            "name": "New Major Issues",
            "description": "New Major issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "168",
            "key": "new_minor_violations",
            "type": "INT",
            "name": "New Minor Issues",
            "description": "New Minor issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "351",
            "key": "new_security_hotspots",
            "type": "INT",
            "name": "New Security Hotspots",
            "description": "New Security Hotspots",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "311",
            "key": "new_vulnerabilities",
            "type": "INT",
            "name": "New Vulnerabilities",
            "description": "New Vulnerabilities",
            "domain": "Security",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "275",
            "key": "open_issues",
            "type": "INT",
            "name": "Open Issues",
            "description": "Open issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "47",
            "key": "quality_profiles",
            "type": "DATA",
            "name": "Profiles",
            "description": "Details of quality profiles used during analysis",
            "domain": "General",
            "direction": 0,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "240",
            "key": "projects",
            "type": "INT",
            "name": "Project branches",
            "description": "Number of project branches",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "58",
            "key": "public_api",
            "type": "INT",
            "name": "Public API",
            "description": "Public API",
            "domain": "Documentation",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "61",
            "key": "public_documented_api_density",
            "type": "PERCENT",
            "name": "Public Documented API (%)",
            "description": "Public documented classes and functions balanced by ncloc",
            "domain": "Documentation",
            "direction": 1,
            "qualitative": True,
            "hidden": True,
            "decimalScale": 1
        },
        {
            "id": "62",
            "key": "public_undocumented_api",
            "type": "INT",
            "name": "Public Undocumented API",
            "description": "Public undocumented classes, functions and variables",
            "domain": "Documentation",
            "direction": -1,
            "qualitative": True,
            "hidden": True
        },
        {
            "id": "294",
            "key": "quality_gate_details",
            "type": "DATA",
            "name": "Quality Gate Details",
            "description": "The project detailed status with regard to its quality gate",
            "domain": "General",
            "direction": 0,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "48",
            "key": "alert_status",
            "type": "LEVEL",
            "name": "Quality Gate Status",
            "description": "The project status with regard to its quality gate.",
            "domain": "Releasability",
            "direction": 1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "315",
            "key": "reliability_rating",
            "type": "RATING",
            "name": "Reliability Rating",
            "description": "Reliability rating",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "342",
            "key": "new_reliability_rating",
            "type": "RATING",
            "name": "Reliability Rating on New Code",
            "description": "Reliability rating on new code",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "313",
            "key": "reliability_remediation_effort",
            "type": "WORK_DUR",
            "name": "Reliability Remediation Effort",
            "description": "Reliability Remediation Effort",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "314",
            "key": "new_reliability_remediation_effort",
            "type": "WORK_DUR",
            "name": "Reliability Remediation Effort on New Code",
            "description": "Reliability remediation effort on new code",
            "domain": "Reliability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "276",
            "key": "reopened_issues",
            "type": "INT",
            "name": "Reopened Issues",
            "description": "Reopened issues",
            "domain": "Issues",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "350",
            "key": "security_hotspots",
            "type": "INT",
            "name": "Security Hotspots",
            "description": "Security Hotspots",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "354",
            "key": "security_hotspots_reviewed",
            "type": "PERCENT",
            "name": "Security Hotspots Reviewed",
            "description": "Percentage of Security Hotspots Reviewed",
            "domain": "SecurityReview",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "355",
            "key": "new_security_hotspots_reviewed",
            "type": "PERCENT",
            "name": "Security Hotspots Reviewed on New Code",
            "description": "Percentage of Security Hotspots Reviewed on New Code",
            "domain": "SecurityReview",
            "direction": 1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        },
        {
            "id": "319",
            "key": "security_rating",
            "type": "RATING",
            "name": "Security Rating",
            "description": "Security rating",
            "domain": "Security",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "343",
            "key": "new_security_rating",
            "type": "RATING",
            "name": "Security Rating on New Code",
            "description": "Security rating on new code",
            "domain": "Security",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "317",
            "key": "security_remediation_effort",
            "type": "WORK_DUR",
            "name": "Security Remediation Effort",
            "description": "Security remediation effort",
            "domain": "Security",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "318",
            "key": "new_security_remediation_effort",
            "type": "WORK_DUR",
            "name": "Security Remediation Effort on New Code",
            "description": "Security remediation effort on new code",
            "domain": "Security",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "352",
            "key": "security_review_rating",
            "type": "RATING",
            "name": "Security Review Rating",
            "description": "Security Review Rating",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "353",
            "key": "new_security_review_rating",
            "type": "RATING",
            "name": "Security Review Rating on New Code",
            "description": "Security Review Rating on New Code",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "356",
            "key": "security_hotspots_reviewed_status",
            "type": "INT",
            "name": "Security Review Reviewed Status",
            "description": "Security Review Reviewed Status",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "358",
            "key": "new_security_hotspots_reviewed_status",
            "type": "INT",
            "name": "Security Review Reviewed Status on New Code",
            "description": "Security Review Reviewed Status on New Code",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "357",
            "key": "security_hotspots_to_review_status",
            "type": "INT",
            "name": "Security Review To Review Status",
            "description": "Security Review To Review Status",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "359",
            "key": "new_security_hotspots_to_review_status",
            "type": "INT",
            "name": "Security Review To Review Status on New Code",
            "description": "Security Review To Review Status on New Code",
            "domain": "SecurityReview",
            "direction": -1,
            "qualitative": False,
            "hidden": True
        },
        {
            "id": "11",
            "key": "skipped_tests",
            "type": "INT",
            "name": "Skipped Unit Tests",
            "description": "Number of skipped unit tests",
            "domain": "Coverage",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "60",
            "key": "statements",
            "type": "INT",
            "name": "Statements",
            "description": "Number of statements",
            "domain": "Size",
            "direction": -1,
            "qualitative": False,
            "hidden": False
        },
        {
            "id": "162",
            "key": "sqale_index",
            "type": "WORK_DUR",
            "name": "Technical Debt",
            "description": "Total effort (in hours) to fix all the issues on the component and therefore to comply to all the requirements.",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False
        },
        {
            "id": "297",
            "key": "sqale_debt_ratio",
            "type": "PERCENT",
            "name": "Technical Debt Ratio",
            "description": "Ratio of the actual technical debt compared to the estimated cost to develop the whole source code from scratch",
            "domain": "Maintainability",
            "direction": -1,
            "qualitative": True,
            "hidden": False,
            "decimalScale": 1
        }
    ],
    "total": 112,
    "p": 1,
    "ps": 100
}



import core.measures_functions as ems_functions
from util.check_functions import (
    check_metric_value,
    check_metric_values,
)


def non_complex_files_density(data_frame):
    """
    Calculates non-complex files density (em1).
    This function calculates non-complex files density measure (em1)
    used to assess the changeability quality subcharacteristic.

    This function gets the dataframe metrics
    and returns the non-complex files density measure (em1).
    """
    files_complexity = data_frame["complexity"]  # m1 metric
    files_functions = data_frame["functions"]  # m2 metric

    check_metric_values(files_complexity, "complexity")
    check_metric_values(files_functions, "functions")

    return ems_functions.calculate_em1(data={
        "complexity": files_complexity,
        "functions": files_functions,
    })


def commented_files_density(data_frame):
    """
    Calculates commented files density (em2).

    This function gets the dataframe metrics
    and returns the commented files density measure (em2).
    """
    files_comment_lines_density = data_frame["comment_lines_density"]  # m4 metric

    check_metric_values(files_comment_lines_density, "comment_lines_density")

    return ems_functions.calculate_em2(data={
        "comment_lines_density": files_comment_lines_density,
    })


def absence_of_duplications(data_frame):
    """
    Calculates duplicated files absence (em3).

    This function gets the dataframe metrics
    and returns the duplicated files absence measure (em3).
    """
    files_duplicated_lines_density = data_frame["duplicated_lines_density"]  # m5 metric

    check_metric_values(files_duplicated_lines_density, "duplicated_lines_density")

    return ems_functions.calculate_em3(data={
        "duplicated_lines_density": files_duplicated_lines_density
    })


def test_coverage(data_frame):
    """
    Calculates test coverage (em6).

    This function gets the dataframe metrics
    and returns the test coverage measure (em6).
    """
    coverage = data_frame["coverage"]  # m6 metric

    check_metric_values(coverage, "coverage")

    return ems_functions.calculate_em6(data={
        "coverage": coverage
    })


def fast_test_builds(data_frame):
    """
    Calculates fast test builds (em5)
    This function gets the dataframe metrics
    and returns the fast test builds measure (em5).
    """
    test_execution_time = data_frame["test_execution_time"]
    tests = data_frame["tests"]

    check_metric_values(test_execution_time, "test_execution_time")
    check_metric_values(tests, "tests")

    return ems_functions.calculate_em5(data={
        "test_execution_time": test_execution_time,
        "tests": tests
    })


def passed_tests(data_frame):
    """
    Calculates passed tests (em4)

    This function gets the dataframe metrics
    and returns the passed tests measure (em4).
    """
    tests = data_frame["tests"]  # m6 metrics
    test_errors = data_frame["test_errors"]  # m7 metrics
    test_failures = data_frame["test_failures"]  # m8 metrics

    check_metric_values(tests, "tests")
    check_metric_value(test_failures, "test_failures")
    check_metric_value(test_errors, "test_errors")

    return ems_functions.calculate_em4(data={
        "tests": tests,
        "test_errors": float(test_errors),
        "test_failures": float(test_failures),
    })



class MeasureSoftGramCoreException(Exception):
    """Base MeasureSoftGram Core exception"""
    pass


class InvalidMetricValue(MeasureSoftGramCoreException):
    """Raised when a invalid metric value is provided to the interpretation function"""
    pass


class InvalidInterpretationFunctionArguments(MeasureSoftGramCoreException):
    """Raised when invalid arguments are provided to an interpretation function"""
    pass


class ValuesAndWeightsOfDifferentSizes(MeasureSoftGramCoreException):
    """Raised when the length of the weight and values are not equal"""
    pass


class ImplicitMetricValueError(MeasureSoftGramCoreException):
    """
    Exception levantanda quando uma métrica não é passada e não é possível
    obter seu valor de maneira implícita.
    """
    pass



# NOTE: Esse dataframe era utilizado nas funções de interpretação,
#       porém foi retirado devido ao empacotamento.
#       Assim, está comentado caso venha a ser utilizado novamente.

# import pandas as pd

# from util.check_functions import check_component_is_valid


# def create_dataframe(measures, components, language_extension):

#     df_columns = measures + ["qualifier"]

#     df = pd.DataFrame(columns=df_columns)

#     for component in components:
#         if check_component_is_valid(component, language_extension):
#             df.at[component["path"], "qualifier"] = component["qualifier"]
#             for metric in component["measures"]:
#                 df.at[component["path"], metric["metric"]] = metric["value"]

#     df.reset_index(inplace=True)
#     return df.rename({"index": "path"}, axis=1)



import math

# import pandas as pd

from util.exceptions import (
    # InvalidInterpretationFunctionArguments,
    InvalidMetricValue,
)


# NOTE: Esses métodos eram utilizados na função do dataframe
#       que era utilizado nas funções de interpretação,
#       porém foi retirado devido ao empacotamento.
#       Assim, estão comentadas caso venham a ser utilizadas novamente.
# def check_component_is_valid(component, language_extension):

#     return (
#         component["qualifier"] == "DIR" or component["language"] == language_extension
#     )


# def check_arguments(data_frame: pd.DataFrame):
#     """
#     Raises an InvalidInterpretationFunctionArguments exception if argument it's not a pandas.DataFrame.
#     """

#     if not isinstance(data_frame, pd.DataFrame):
#         raise InvalidInterpretationFunctionArguments(
#             "Expected data_frame to be a pandas.DataFrame"
#         )


# def check_number_of_files(number_of_files):
#     """
#     Raises an InvalidMetricValue exception if the number of files is lesser or equal than 0.
#     """

#     if number_of_files <= 0:
#         raise InvalidMetricValue("The number of files is lesser or equal than 0")


def check_metric_value(metric_value, metric):
    try:
        if metric_value is None or math.isnan(float(metric_value)):
            raise InvalidMetricValue(f'"{metric}" has an invalid metric value')
    except (ValueError, TypeError):
        raise InvalidMetricValue(f'"{metric}" has an invalid metric value')


def check_metric_values(metric_values, metric):
    if not isinstance(metric_values, list):
        raise InvalidMetricValue(f'"{metric}" is not a list')

    if not len(metric_values):
        raise InvalidMetricValue(f'"{metric}" is empty')

    for value in metric_values:
        check_metric_value(value, metric)



import numpy as np

from util.exceptions import ValuesAndWeightsOfDifferentSizes


def weighting_operation(values, weights):
    if len(values) != len(weights):
        raise ValuesAndWeightsOfDifferentSizes(
            "The length of weight and values are not equal",
        )

    values = np.array(values)
    weights = np.array(weights)

    return values * weights



SONARQUBE_SUPPORTED_MEASURES = [
    {
        "passed_tests": {
            "metrics": [
                "tests",
                "test_failures",
                "test_errors",
            ],
        }
    },
    {
        "test_builds": {
            "metrics": [
                "test_execution_time",
                "tests",
            ],
        }
    },
    {
        "test_coverage": {
            "metrics": [
                "coverage",
            ],
        }
    },
    {
        "non_complex_file_density": {
            "metrics": [
                "functions",
                "complexity",
            ],
        }
    },
    {
        "commented_file_density": {
            "metrics": [
                "comment_lines_density",
            ],
        }
    },
    {
        "duplication_absense": {
            "metrics": [
                "duplicated_lines_density",
            ],
        }
    },
    # {
    #     "ci_feedback_time": {
    #         "metrics": [
    #             "number_of_build_pipelines_in_the_last_x_days",
    #             "runtime_sum_of_build_pipelines_in_the_last_x_days",
    #         ],
    #     }
    # },
    # {
    #     "team_throughput": {
    #         "metrics": [
    #             "number_of_resolved_issues_with_US_label_in_the_last_x_days",
    #             "total_number_of_issues_with_US_label_in_the_last_x_days",
    #         ],
    #     }
    # },
]



import requests
from marshmallow.exceptions import ValidationError

from core.schemas import (
    CalculateMeasureSchema,
    CalculateSubCharacteristicSchema,
    CalculateCharacteristicSchema,
    CalculateSQCSchema
)
from core.agregation import aggregation_operation
from core.weighting import weighting_operation
from util.constants import MEASURES_INTERPRETATION_MAPPING


def calculate_aggregated_value(values_list, weights_list):
    weighted_items = weighting_operation(values_list, weights_list)
    return aggregation_operation(weighted_items, weights_list)


def calculate_measures(extracted_measures):
    # Validate if outter keys is valid
    try:
        data = CalculateMeasureSchema().load(extracted_measures)
    except ValidationError as error:
        return {
            "error": "Failed to validate request",
            "schema_errors": error.messages,
            "code": requests.codes.unprocessable_entity
        }

    # Objeto retornado em caso de sucesso
    response_data = {"measures": []}

    valid_measures = MEASURES_INTERPRETATION_MAPPING.keys()

    for measure in data["measures"]:
        measure_key: str = measure['key']

        if measure_key not in valid_measures:
            return {
                "error": f"Measure {measure_key} is not supported",
                "code": requests.codes.unprocessable_entity
            }

        measure_params = measure["parameters"]
        schema = MEASURES_INTERPRETATION_MAPPING[measure_key]["schema"]

        try:
            validated_params = schema().load(measure_params)
        except ValidationError as exc:
            return {
                "error": f"Metric parameters {measure_key} are not valid",
                "schema_errors": exc.messages,
                "code": requests.codes.unprocessable_entity
            }

        interpretation_function = MEASURES_INTERPRETATION_MAPPING[measure_key]["interpretation_function"]
        result = interpretation_function(validated_params)

        response_data["measures"].append({
            "key": measure_key,
            "value": result,
        })

    return response_data


def calculate_subcharacteristics(extracted_subcharacteristics):
    try:
        data = CalculateSubCharacteristicSchema().load(extracted_subcharacteristics)
    except ValidationError as error:
        return {
            "error": "Failed to validate request",
            "schema_errors": error.messages,
            "code": requests.codes.unprocessable_entity
        }

    response_data = {"subcharacteristics": []}

    for subcharacteristic in data["subcharacteristics"]:
        subcharacteristic_key: str = subcharacteristic["key"]

        values_list, weights_list = [], []
        for measure in subcharacteristic["measures"]:
            values_list.append(measure["value"])
            weights_list.append(measure["weight"])

        aggregated_value = calculate_aggregated_value(values_list, weights_list)

        response_data["subcharacteristics"].append({
            "key": subcharacteristic_key,
            "value": aggregated_value,
        })

    return response_data


def calculate_characteristics(extracted_characteristics):
    try:
        data = CalculateCharacteristicSchema().load(extracted_characteristics)
    except ValidationError as error:
        return {
            "error": "Failed to validate request",
            "schema_errors": error.messages,
            "code": requests.codes.unprocessable_entity
        }

    response_data = {"characteristics": []}

    for characteristic in data["characteristics"]:
        characteristic_key: str = characteristic["key"]

        values_list, weights_list = [], []
        for measure in characteristic["subcharacteristics"]:
            values_list.append(measure["value"])
            weights_list.append(measure["weight"])

        aggregated_value = calculate_aggregated_value(values_list, weights_list)

        response_data["characteristics"].append({
            "key": characteristic_key,
            "value": aggregated_value,
        })

    return response_data


def calculate_sqc(extracted_sqc):
    try:
        data = CalculateSQCSchema().load(extracted_sqc)
    except ValidationError as error:
        return {
            "error": "Failed to validate request",
            "schema_errors": error.messages,
            "code": requests.codes.unprocessable_entity
        }

    response_data = {"sqc": []}

    sqc = data["sqc"]
    sqc_key: str = sqc["key"]

    values_list, weights_list = [], []
    for characteristic in sqc["characteristics"]:
        values_list.append(characteristic["value"])
        weights_list.append(characteristic["weight"])

    aggregated_value = calculate_aggregated_value(values_list, weights_list)

    response_data["sqc"].append({
        "key": sqc_key,
        "value": aggregated_value,
    })

    return response_data



from .sonarqube_available_metrics import SONARQUBE_AVAILABLE_METRICS  # noqa: F401
from .sonarqube_supported_metrics import SONARQUBE_SUPPORTED_MEASURES  # noqa: F401
from .default_pre_config import DEFAULT_PRE_CONFIG  # noqa: F401



import os

import requests

from staticfiles import SONARQUBE_AVAILABLE_METRICS, SONARQUBE_SUPPORTED_MEASURES


class Sonarqube:

    def __init__(self):
        self.endpoint = os.getenv("SONAR_URL", "https://sonarcloud.io/api/metrics/search")

    def extract_supported_metrics(self, metrics):
        try:
            request = requests.get(self.endpoint)

            if request.ok:
                data = request.json()
            else:
                data = SONARQUBE_AVAILABLE_METRICS
        except Exception:
            data = SONARQUBE_AVAILABLE_METRICS

        return self.__extract_sonarqube_supported_metrics(metrics, data)

    def __extract_sonarqube_supported_metrics(self, metrics, sonar_metrics):
        collected_metrics = {}
        supported_metrics = []

        # NOTE: list comprehension que preenche o supported_metrics apenas
        #       com os valores das "metrics" do dicionário de SUPPORTEDs
        [
            supported_metrics.extend(list(x.values())[0]['metrics'])
            for x in SONARQUBE_SUPPORTED_MEASURES
        ]

        for component in metrics:
            qualifier = component['qualifier']
            path = component['path' if qualifier != 'TRK' else 'name']
            collected_metrics[path] = {
                'qualifier': qualifier,
                'measures': []
            }

            for obj in component['measures']:
                metric_key = obj['metric']
                if metric_key not in supported_metrics:
                    continue

                collected_metrics[path]['measures'].append({
                    'metric': metric_key,
                    'value': float(obj['value'])
                })

        return collected_metrics



# flake8: noqa
# pylint: skip-file
DEFAULT_PRE_CONFIG = {
  "characteristics": [
    {
      "key": "reliability",
      "weight": 50,
      "subcharacteristics": [
        {
          "key": "testing_status",
          "weight": 100,
          "measures": [
            {
              "key": "passed_tests",
              "weight": 33
            },
            {
              "key": "test_builds",
              "weight": 33
            },
            {
              "key": "test_coverage",
              "weight": 34
            }
          ]
        }
      ]
    },
    {
      "key": "maintainability",
      "weight": 50,
      "subcharacteristics": [
        {
          "key": "modifiability",
          "weight": 100,
          "measures": [
            {
              "key": "non_complex_file_density",
              "weight": 33
            },
            {
              "key": "commented_file_density",
              "weight": 33
            },
            {
              "key": "duplication_absense",
              "weight": 34
            }
          ]
        }
      ]
    },
    # {
    #   "key": "functional_suitability",
    #   "weight": 34.0,
    #   "subcharacteristics": [
    #     {
    #       "key": "functional_completeness",
    #       "weight": 100.0,
    #       "measures": [
    #         {
    #           "key": "team_throughput",
    #           "weight": 100.0
    #         }
    #       ]
    #     }
    #   ]
    # }
  ]
}


import numpy as np
import pandas as pd


def create_coordinate_pair(min_threshhold, max_threshold, reverse_y=False):
    """
    Creates a pair of values (x, y), based on the min and max thresholds.
    """
    y = np.array([0, 1]) if reverse_y else np.array([1, 0])

    return np.array([min_threshhold, max_threshold]), y


def get_files_data_frame(data_frame: pd.DataFrame) -> pd.DataFrame:
    """
    Returns a data frame with files data.
    """

    return data_frame[
        (data_frame["qualifier"] == "FIL") & (data_frame["ncloc"].astype(float) > 0)
    ]


def get_test_root_dir(data_frame: pd.DataFrame) -> pd.Series:
    """
    Returns a Series populated with "tests" index line from the data_frame.
    """
    dirs = data_frame[data_frame["qualifier"] == "DIR"]

    return dirs.loc[dirs["tests"].astype(float).idxmax()]



import numpy as np


def aggregation_operation(weighted_value, weights):
    weighted_value = np.array(weighted_value)
    weights = np.array(weights)
    normalized_value = np.linalg.norm(weighted_value)
    max_value = np.linalg.norm(weights)
    return normalized_value / max_value
